# count
OUTS_DIR = "outs"
RAW_MATRIX_DIR_SUFFIX = "raw"
FILTERED_MATRIX_DIR_SUFFIX = "filtered"
MATRIX_FILE_NAME = "matrix.mtx.gz"
FEATURE_FILE_NAME = "features.tsv.gz"
BARCODE_FILE_NAME = "barcodes.tsv.gz"
STAR_BAM_SUFFIX = "Aligned.out.bam"
TAG_BAM_SUFFIX = "aligned_posSorted_addTag.bam"
STARSOLO_BAM_SUFFIX = "Aligned.sortedByCoord.out.bam"
COUNTS_FILE_NAME = "counts.tsv"

# mkref
GENOME_CONFIG = "celescope_genome.config"
