{% include "html/utils/start.html"%}

    {% if demultiplexing_summary is defined %}
    {% set step_summary = demultiplexing_summary %}
    {% include "html/utils/step.html"%}
    {% endif %}

    {% if table_probe is defined %}
    {% set table = table_probe %}
    {% include "html/utils/table.html"%}
    {% endif %}

    {% if table_primer is defined %}
    {% set table = table_primer %}
    {% include "html/utils/table.html"%}
    {% endif %}

{% include "html/utils/end.html"%}

