{% include "html/utils/start.html"%}

    {% if demultiplexing_summary is defined %}
    {% set step_summary = demultiplexing_summary %}
    {% include "html/utils/step.html"%}
    {% endif %}

    {% if mapping_summary is defined %}
    {% set step_summary = mapping_summary %}
    {% include "html/utils/step.html" %}
    {% endif %}

    {% if cells_summary is defined%}
    {% set step_summary = cells_summary %}
    {% include "html/bulk_rna/cells.html" %}
    {% endif %}


{% include "html/utils/end.html"%}

