{% if step_summary.table_dict is defined %}

    {% if step_summary.table_dict.id == "marker_genes" %}
        {% include "./html/utils/table_script/rna.html" %}
    {% elif step_summary.table_dict.id == "clonetypes" %}
        {% include "./html/utils/table_script/vdj.html" %}
    {% elif step_summary.table_dict.id == "variant" %}
        {% include "./html/utils/table_script/snp.html" %}
    {% elif step_summary.table_dict.id == "bulk_vdj" %}
        {% include "./html/utils/table_script/bulk_vdj.html" %}
    {% elif step_summary.table_dict.id == "bulk_rna" %}
        {% include "./html/utils/table_script/bulk_rna.html" %}
        {% elif step_summary.table_dict.id == "pathseq" %}
        {% include "./html/utils/table_script/pathseq.html" %}
    {% endif %}

    <h3>{{step_summary.table_dict['title']}}</h3>
    <div style=" margin-left: 3%; margin-right:3%;">
        {{step_summary.table_dict['table'] | safe}}
    </div>

{% endif %}